/*
   AngelCode Windows Wrapper Library
   Copyright (c) 2004-2005 Andreas Jnsson
  
   This software is provided 'as-is', without any express or implied 
   warranty. In no event will the authors be held liable for any 
   damages arising from the use of this software.

   Permission is granted to anyone to use this software for any 
   purpose, including commercial applications, and to alter it and 
   redistribute it freely, subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you 
      must not claim that you wrote the original software. If you use
	  this software in a product, an acknowledgment in the product 
	  documentation would be appreciated but is not required.

   2. Altered source versions must be plainly marked as such, and 
      must not be misrepresented as being the original software.

   3. This notice may not be removed or altered from any source 
      distribution.
  
   Andreas Jnsson
   andreas@angelcode.com
*/


#include "ac_window.h"
#include <winuser.h>

//=============================================================================

acCWindow::acCWindow()
{
	hWnd   = 0;
	hAccel = 0;
	hMenu  = 0;

	originalProc = 0;
}

acCWindow::~acCWindow()
{
	if( hWnd && IsWindow(hWnd) )
		DestroyWindow(hWnd);

	if( hMenu ) 
		DestroyMenu(hMenu);
}

//=============================================================================

// static
acCWindow *acCWindow::FromHandle(HWND hWnd)
{
	acCWindow *wnd;
	wnd = (acCWindow *)GetWindowLong(hWnd, GWL_USERDATA);
	return wnd;
}

//=============================================================================

int acCWindow::Create(const char *title, int width, int height, DWORD style, DWORD styleEx, acCWindow *parent, const char *className)
{
	if( className == 0 )
	{
		// Register a default window class
		int r = RegisterClass("Basic Window", 0, AC_REGDEFBRUSH, AC_REGDEFICON, AC_REGDEFICON, AC_REGDEFCURSOR);
		if( FAILED(r) )
			return r;

		className = "Basic Window";
	}

	HWND hParent = 0;
	if( parent ) hParent = parent->hWnd;

	HINSTANCE hInst = GetModuleHandle(0);

    // Create the window
	// We set up a hook procedure to connect our class instance with the window handle
	HookCreate(this);
    HWND hWndNew = CreateWindowEx( styleEx, className, 
		                           title, style, CW_USEDEFAULT, CW_USEDEFAULT,
						           width,height,hParent,NULL,hInst,0 ); 
	if( hWndNew == NULL )
		return EWND_CREATE_WINDOW_FAILED;
	
	return 0;
}

//=============================================================================

int acCWindow::SetAccelerator(const char *accel)
{
	// Get the module handle
	HINSTANCE hInst = GetModuleHandle(0);
	
	// Load the accelerator table
	if( accel != 0 )
	{
		hAccel = LoadAccelerators(hInst, accel);
		if( hAccel == 0 )
			return EWND_LOAD_RESOURCE_FAILED;
	}
	else
		hAccel = 0;

	// Note that accelerator tables loaded from resources
	// are automatically freed at termination

	return 0;
}

//=============================================================================

int acCWindow::SetMenu(const char *menu)
{
	// Get the module handle
	HINSTANCE hInst = GetModuleHandle(0);

	// Destroy old menu
	if( hMenu != 0 )
	{
		::SetMenu(hWnd, 0);
		DestroyMenu(hMenu);
	}
	hMenu = 0;

	// Load the menu
	if( menu != 0 )
	{
		hMenu = LoadMenu(hInst, menu);
		if( hMenu == 0 )
			return EWND_LOAD_RESOURCE_FAILED;

		::SetMenu(hWnd, hMenu);
	}

	return 0;
}

//=============================================================================

// static
int acCWindow::RegisterClass(const char *className, UINT style, HBRUSH newBgBrush, HICON newIcon, HICON newSmallIcon,
						   HCURSOR newCursor)
{
	// Get the module handle
	HINSTANCE hInst = GetModuleHandle(0);

	// See if the class has been registered before
	WNDCLASSEX wndClass;
	if( GetClassInfoEx(hInst, className, &wndClass) )
		return 0;

	// Register the new class
	HBRUSH  hBgBrush   = newBgBrush;
	HICON   hIcon      = newIcon;
	HCURSOR hCursor    = newCursor;
	HICON   hIconSmall = newSmallIcon;

	if( int(hIcon)      == -1 ) hIcon      = LoadIcon(NULL, IDI_APPLICATION);
	if( int(hIconSmall) == -1 ) hIconSmall = 0;
	if( int(hCursor)    == -1 ) hCursor    = LoadCursor(NULL, IDC_ARROW);
	if( int(hBgBrush)   == -1 ) hBgBrush   = (HBRUSH)::GetStockObject(BLACK_BRUSH);

	// Register the window class
	wndClass.cbSize        = sizeof(WNDCLASSEX);
	wndClass.style         = style;
	wndClass.lpfnWndProc   = (WNDPROC)WndProc;
	wndClass.cbClsExtra    = 0;
	wndClass.cbWndExtra    = 0;
	wndClass.hInstance     = hInst;
	wndClass.hIcon         = hIcon;
	wndClass.hCursor       = hCursor;
	wndClass.hbrBackground = hBgBrush;
	wndClass.lpszMenuName  = 0;
	wndClass.lpszClassName = className;
	wndClass.hIconSm       = hIconSmall;

	ATOM windowClass = RegisterClassEx( &wndClass );
	if( !windowClass )
		return EWND_REGISTER_CLASS_FAILED;

	return 0;
}

//=============================================================================

int acCWindow::Attach(HWND newHWnd)
{
	if( hWnd ) return EWND_ALREADY_ATTACHED;
	if( FromHandle(newHWnd) ) return EWND_OTHER_IS_ALREADY_ATTACHED;

	hWnd = newHWnd;
	SetWindowLong(hWnd, GWL_USERDATA, (LONG)this);

	return 0;
}

int acCWindow::Subclass(HWND newHWnd)
{
	if( hWnd ) return EWND_ALREADY_ATTACHED;
	if( FromHandle(newHWnd) ) return EWND_OTHER_IS_ALREADY_ATTACHED;

	originalProc = (WNDPROC)SetWindowLong(newHWnd, GWL_WNDPROC, (LONG)WndProc);

	Attach(newHWnd);

	return 0;
}

int acCWindow::Detach()
{
	if( hWnd == 0 ) return EWND_NOT_ATTACHED;

	SetWindowLong(hWnd, GWL_USERDATA, 0);

	if( originalProc ) 
		SetWindowLong(hWnd, GWL_WNDPROC, (LONG)originalProc);

	hWnd = 0;

	return 0;
}

//=============================================================================

// static
// If wait is true then the function will wait until a message arrives.
// The function returns true if it receives WM_QUIT.
bool acCWindow::CheckMessages(bool wait)
{
	if( wait )
		WaitMessage();

    MSG msg;
	while( PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
    {
		if( msg.message == WM_QUIT )
			return true;

		// Let window and parents translate the message
		HWND hWnd = msg.hwnd;
		bool isTranslated = false;
		while( hWnd )
		{
			acCWindow *wnd = (acCWindow *)FromHandle(hWnd);
			if( wnd )
			{
				if( isTranslated = wnd->TranslateMessage(&msg) ) 
					break;
			}

			hWnd = GetParent(hWnd);
		}

		if( !isTranslated )
		{
			// Translate key messages to character messages
			::TranslateMessage(&msg);

			// Send the messages to the window procedure
			DispatchMessage(&msg);
		}
	}

    return false;
}

//=============================================================================

bool acCWindow::TranslateMessage(MSG *msg)
{
	if( hAccel && TranslateAccelerator(hWnd, hAccel, msg) )
		return true;

	return false;
}

//=============================================================================

HWND acCWindow::GetHandle()
{
	return hWnd;
}

void acCWindow::Invalidate(BOOL erase)
{
	InvalidateRect(hWnd, 0, erase);
}

//=============================================================================

LRESULT acCWindow::DefWndProc(UINT msg, WPARAM wParam, LPARAM lParam)
{
	if( originalProc )
		return CallWindowProc(originalProc, hWnd, msg, wParam, lParam);

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

LRESULT acCWindow::MsgProc(UINT msg, WPARAM wParam, LPARAM lParam)
{
	// Let the default window procedure handle 
	// any message that we don't care about
	return DefWndProc(msg, wParam, lParam);
}

// static
// Message handler which passes messages to the application class.
LRESULT CALLBACK acCWindow::WndProc(HWND hWnd, UINT msg, 
								   WPARAM wParam, LPARAM lParam)
{
	acCWindow *wnd;

	// Extract the object from the window handle.
	wnd = (acCWindow *)FromHandle(hWnd);
	if( wnd )
	{
		// Let the object handle it.
		return wnd->MsgProc(msg, wParam, lParam);
	}

	// If no class instance is found we call the default procedure
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

//=============================================================================

// TODO: Make thread safe
HHOOK      acCWindow::hCreateHook = 0;
acCWindow *acCWindow::wndCreator  = 0;

// static
int acCWindow::HookCreate(acCWindow *wnd)
{
	if( wnd != 0 )
	{
		wndCreator  = wnd;
		hCreateHook = SetWindowsHookEx(WH_CBT, CreateProc, 0, ::GetCurrentThreadId());
		if( hCreateHook == 0 )
		{
			// Failed, use GetLastError() to know what happened
			return EWND_HOOK_FAILED;
		}
	}
	else
	{
		if( hCreateHook && !UnhookWindowsHookEx(hCreateHook) )
		{
			// Failed, use GetLastError() to know what happened
			return EWND_HOOK_FAILED;
		}
		hCreateHook = 0;
		wndCreator  = 0;
	}

	return 0;
}

// static
LRESULT CALLBACK acCWindow::CreateProc(int nCode, WPARAM wParam, LPARAM lParam)
{
	if( nCode == HCBT_CREATEWND )
	{
		// Connect the new window with the class instance
		HWND hWnd = (HWND)wParam;
		CBT_CREATEWND *cw = (CBT_CREATEWND*)lParam;

		if( wndCreator )
			wndCreator->Attach(hWnd);

		// Unregister the hook procedure here
		HookCreate(0);

		// Return 0 to allow continued creation
		return 0;
	}

	return CallNextHookEx(hCreateHook, nCode, wParam, lParam);
}

//=============================================================================

// WM_DRAWITEM is sent to the parent for owner drawn 
// controls and the parent calls this method on the child.
// Return true, if the message is handled
BOOL acCWindow::DrawItem(DRAWITEMSTRUCT *di)
{
	return FALSE;
}

//=============================================================================

bool acCWindow::IsVisible()
{
	LONG style = GetWindowLong(hWnd, GWL_STYLE);
	if( style & WS_VISIBLE )
		return true;

	return false;
}

//=============================================================================

void acCWindow::UpdateWindowText(const char *text)
{
	int len = GetWindowTextLength(hWnd);
	char *str = new char[len+1];
	
	GetWindowText(hWnd, str, len);
	if( strcmp(text, str) != 0 )
		SetWindowText(hWnd, text);

	delete[] str;
}