#include <stdlib.h>
#include <errno.h>
#include <vector>

#include "image.h"
#include "pngfile.h"
#include "..\libraries\png.h"
#include "file.h"
using namespace std;

int writePNG(const char *filename, CImage *img, int saveChannelsMode, bool saveInterlaced)
{
	CFile file;
	png_structp png;
	png_infop   info;

	// Open the file
	if( file.Open(filename, "wb") < 0 )
	{
		if( errno == EACCES )
			return -1;
		if( errno == EEXIST )
			return -2;
		if( errno == EINVAL )
			return -3;
		return -1;
	}

	// Initialize the png structure with no special error handling
	png = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
	if( png == 0 )
	{
		file.Close();
		return -1;
	}

	// Allocate and initialize the image information data
	info = png_create_info_struct(png);
	if(info == NULL)
	{
		file.Close();
		png_destroy_write_struct(&png, png_infopp_NULL);
		return -1;
	}

	// Set error handling.  
	if( setjmp(png_jmpbuf(png)) )
	{
		// If we get here, we had a problem
		file.Close();
		png_destroy_write_struct(&png, &info);
		return -1;
	}

	// Set the file stream for writing
	png_init_io(png, file.file);

	// Set the image information
	int color_type = PNG_COLOR_TYPE_RGB;
	if( saveChannelsMode == 1 ) color_type = PNG_COLOR_TYPE_RGB_ALPHA;
	if( saveChannelsMode == 2 ) color_type = PNG_COLOR_TYPE_GRAY;
	int interlace_type = saveInterlaced ? PNG_INTERLACE_ADAM7 : PNG_INTERLACE_NONE;
	png_set_IHDR(png, info, img->width, img->height, 8, color_type, 
		interlace_type, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	// Write the file header information
	png_write_info(png, info);

	// We need to swap the order of the color channels (not alpha)
	if( saveChannelsMode != 2 )
		png_set_bgr(png);

	// Remove alpha channel as specified
	if( saveChannelsMode == 0 ) 
		png_set_filler(png, 0, PNG_FILLER_AFTER);

	// Set up an array of pointer to the pixel rows
	png_bytep *rows = new png_bytep[img->height];
	if( rows == 0 )
	{
		file.Close();
		png_destroy_write_struct(&png, &info);
		return -1;
	}

	// If we are only saving the alpha channel we must convert the image to alpha only
	vector<BYTE> alphaImage;
	if( saveChannelsMode == 2 )
	{
		alphaImage.resize(img->width*img->height, false);
		for( int n = 0; n < img->height*img->width; n++ )
			alphaImage[n] = BYTE(img->pixels[n]>>24);

		for( n = 0; n < img->height; n++ )
			rows[n] = png_bytep(&alphaImage[0] + n*img->width);
	}
	else
	{
		// Simply take pointers to the original image
		for( int n = 0; n < img->height; n++ )
			rows[n] = png_bytep(img->pixels + n*img->width);
	}

	// Write the image
	png_write_image(png, rows);

	// Complete the file writing
	png_write_end(png, info);

	// Clean up
	delete[] rows;

	png_destroy_write_struct(&png, &info);

	file.Close();

	return 0;
}

